/* dirx_256.c : module de base pour WinXP+DirectX ou NT 4
Description des instructions de ce module dans doc*.txt.
Guillaume LAMONOCA
Supelec (94)

Dreaming of making yourself WinDoom ? This module is for you !
(Read doc.txt and use the loackscreen() function - to retrieve 
 the "video ram" pointer -, setbigpalette(),refresh(),swap())

Module specialement dedie au fonctionnement en double buffering
avec swapping d'adresse video : recherche de vitesse maximale.

Module fonctionnant avec Direct X (extensions gratuite de Win95)
 - en 16 couleurs (GLCB normal) en mode 256 couleurs physiques
 - en 256 couleurs (GLCB+EXTENSION) en mode 256 couleurs physiques
DirectX est utilis en mode plein cran.

  Ce module n'utilise pas les fonctions GDI pour le graphisme.
Avantages :
        - pas de substition de couleur silencieuse
        comme c'est le cas lors de l'utilisation
        du GDI; on controle completement les couleurs. 
        - les blocs graphiques avec silhouette (sprites)
        consomme la quantite minimale de memoire
        (masque stoque avec le bloc graphique :
         la couleur 255 est la couleur transparente)
Inconvenients:
        - vous n'avez que 254 couleurs (evitez la couleur 0
        et la couleur 255; si vous le pouvez evitez les
        16 premieres couleurs pour le portage vers GEM)
        La couleur 255 est la transparence des sprites
        - les fonctions graphiques ne sont pas tres rapides
        (ex: polyfill) car ce sont des routines en C
        Ce modules est plutot destine a vous permettre
        d'attaquer directement la mmoire cran
        grace a lockworkscreen() et releaseworkscreen(), et
        donc d'optimiser vous meme vos routines.

En d'autres termes, si votre programme a pour vocation de
manipuler peu de bitmaps, ou si vous modifiez rarement
la palette de couleurs, mieux vaut utiliser wing_gdi.c (module
GLCB pour WinG, mais qui utilise les fonctions GDI et consomme
le double de la memoire vraiment necessaire pour les blocs
graphiques avec silhouette).

Attention le debuggage est difficile avec ce module (le contrle
total de l'ecran empche la possibilit de tracer).
Je vous conseille de developper avec le module WinG (traable)
en mode 64K (pas de risque de perte de palette) puis de passer
au module DirectX sans changer vos sources.

DirectX ajoute a Windows des fonctions permettant d'afficher
des bitmaps a tres tres grande vitesse. (WinDoom utilise
DirectX). De plus DirectX permet de travailler directement 
dans la memoire video et dans un format plus que sympatique.
DirectX permet de prendre le contrle total de l'cran.

Utiliser Microsoft Visual C++ 4.0

Si vous obtenez de nombreux Warnings, plutot que de
corriger vos sources, n'hesitez pas a desactiver ces
warnings! 

Marche a suivre pour produire exemple.exe
-----------------------------------------

Mettez les fichiers suivants dans un projet exemple.mak :

dirx_256.c                                
ddraw.lib    (indispensable pour DirectX fourni avec Game SDK)
dsound.lib   (indispensable pour DirectX fourni avec Game SDK)
exemple.c    (ou vos sources)
winmm.lib

Remarques
---------

Il faut absolument eviter de sortir brusquement de
votre programme car sous windows les zones memoires allouees
dynamiquement doivent etre liberees a la fin de votre
programme! Une interruption brusque laissera ces zones
memoires inutilisables pour le systeme! Donc prevoyez
toujours un moyen de sortir de maniere normale de votre
programme (liberation des objets, des zones memoires et
appel a la fonction killsystem()).
(Ou alors ajouter un appel vers votre fonction free_all()
 a la fin de la fonction WinMain dans ce module en
 definissant la constante USER_FREE_FUNCTION)

Pour les samples, vous devez avoir installe un driver
correct avec le panneau de configuration de windows.
De plus il ne faut pas jouer un sample a une frequence
trop haute ou trop basse (ideal:11025 Hz).



Interet principal de DirectX:
----------------------------

Grace a la valeur retournee par lockworkscreen() - l'adresse ecran -
vous pouvez travailler directement dans les bitmaps offscreen
(ca simule des rams video en quelque sorte...) lorsque vous tes
en mode fentre ou mieux dans la ram video en mode plein cran.
Format des "ram video": 1 octet= 1 pixel, attention 0 donnera
toujours du noir et 255 toujours du blanc (seul le GDI vous
donnera l'illusion d'avoir vraiment 256 couleurs; cf wing_gdi.c).

Mais le plus formidable dans DirectX est la possibilit de
choisir un affichage plein cran, o le rafrachissement du
contenu de la fentre (de l'cran en fait) n'est plus assure
par une copie de pixel mais par un simple changement d'adresse
vido (donc beaucoup plus rapide).


Amusez vous bien... Gardez en tete que WinDoom utilise la
meme technique que ce module : controle des couleurs (fonction
setbigpalette), ordre de rafraichissement (refresh ou swap)
et acces direct aux bitmaps offscreens ou memoire video par le 
biais de petites routines assembleurs hyper optimisees... 
Ajoutez le fait que grace a Windows vous avez facilement acces 
a la memoire etendue (et meme virtuelle) et a l'execution de sons
digitalises et de musiques MIDI, et vous obtenez une methode 
accessible aux debutants en C, qui permet neanmoins de realiser 
des logiciels qui n'ont rien a envier aux super-productions 
(avec la gestion specifiques des cartes hardware en moins! 
 Quel luxe!).

La constante RATIO peut vous permettre de faire
un zoom d'un facteur 2 si vous vous contentez d'un faible
resolution (320x200) comme c'est le cas par defaut.
La perte de temps de ce zoom peut etre negligeable par rapport
au temps de calcul d'une image


Guillaume LAMONOCA
www.minilgos.com

PS:Si vous faites beaucoup de profit avec ce module, SVP, pensez
   a moi! Les temps sont durs et un pourcentage meme symbolique
   serait fort apprecie... (ceci dit, tous les modules GLCB
   sont freeware a la diffusion comme a l'utilisation : vous
   n'etes pas tenu de reverser la moindre royalty... la seule
   chose que je vous demande, c'est de profiter de l'existence
   des autres modules GLCB pour realiser des versions pour
   les autres machines afin de les aider a survivre)
   'gem_wing.c' est un module pour GEM destine specialement
   pour des conversions rapides de programmes DirectX vers GEM.

   Symbolic royalty or donations expected. But GLCB is freeware,
   you have no royalties to pay back. I just want you to compile
   your program for other platforms (in order to help them to
   survive). Since other GLCB modules exist, use them to provide
   other platform versions of your programs...
*/




#include <windows.h>
#include <windowsx.h>
#include <mmsystem.h>

#include "ddraw.h"
#include "dsound.h"

#include <process.h>             
#ifdef BORLANDC
#include <alloc.h>
#else
#include <malloc.h>
#endif
#include "header.h"
#include <stdlib.h>
#include <setjmp.h>
#include <string.h>


#undef huge
#define huge /**/






/********************************************************************/
/* variables globales */

/* facteur de grossissement (selon mode graphique physique)*/
         



/* modes autoriss sur une carte graphique classique 1Mo de RAM: */
/* 320*200 = (WIDTH*RATIO)*(HEIGHT*RATIO) (Win9x only) */
/* 320*240 = (WIDTH*RATIO)*(HEIGHT*RATIO) (Win9x only) */
/* 640*400 = (WIDTH*RATIO)*(HEIGHT*RATIO) */
/* 640*480 = (WIDTH*RATIO)*(HEIGHT*RATIO) */
/* 800*600 = (WIDTH*RATIO)*(HEIGHT*RATIO) */

#define WIDTH 640
#define HEIGHT 400

/* valeurs autorises : 1, 2 */
#define RATIO 1

unsigned long	BPL=WIDTH*RATIO; //will be adjusted in lockworkscreen()

int video_offset_not_null=0;

#define NOM_FENETRE "Nom fentre"
#define NOM_CLASSE "OutputWClass"

/*#define USER_FREE_FUNCTION*/
/* pour provoquer l'appel a votre fonction de liberation
   memoire void free_all(void) lors d'un arret brutal    */


#define ONLY_ONCE
           /* un seul exemplaire du programme   */
           /* en cours d'execution              */



static  jmp_buf   jmpb;


static unsigned int oldpal[16]=
{
    0x0000,0x1008,0x3080,0x5088,
    0x2800,0x3808,0x4880,0x5ccc,
    0x3888,0x400f,0x50f0,0x60ff,
    0x4f00,0x5f0f,0x7ff0,0x8fff
};

#ifdef EXTENSION
static unsigned int oldpal256[256*3];
#endif

               

static void *cursorbob=NULL;
static void *cursorbak=NULL;
static HCURSOR visible_cursor;
static HCURSOR invisible_cursor;
                              
static long randstamp;
#define MASKSAMPLE 0xff
#define MASKRANDOM 0x000fffffL
#define MAXSAMPLE (MASKSAMPLE+1)
static long splfrq[MAXSAMPLE]={0};
static long spllen[MAXSAMPLE]={0};
                        
static HANDLE hInst;
static HWND window;
static HBITMAP holdbitmap;
static PAINTSTRUCT ps;
static int nbscreen=2;
static int nocolor=15;
static int cur_color;

static int xcur=0;
static int ycur=0;

static int realmsex=WIDTH/2-8;
static int realmsey=HEIGHT/2-8;
static int realmsemode=1;
int mousex,mousey,mousek,vblclock;
int msex,msey,msek,mclk;

static int filltab[HEIGHT];

static int nbrbuffer=0;
static int lstbufptr=0;
static int msebufptr=0;
static int msebuffer[128];

char keymap[256]={
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
};

char keystate[256]={
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
};


static int keybuffer[256];
static int keybufptr=0;
static int keybufend=0;
static int keybufnbr=0;





/********************************************************************/
/* fontes */

static unsigned int sysfonte[64*16]={
0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0xc3,0xc3,0xc3,0xc3,0xc7,0xe7,0xc7,0xe7,0x1c,0x1c,0x1c,0x18,0x18,0x00,0x18,0x00,
0x81,0x81,0x81,0xdb,0xff,0xff,0xff,0xff,0x36,0x36,0x24,0x00,0x00,0x00,0x00,0x00,
0xc8,0x88,0x00,0x81,0x80,0x00,0x81,0xdb,0x13,0x33,0x7e,0x36,0x37,0x7e,0x24,0x00,
0x83,0x01,0x01,0x81,0xc0,0x00,0x80,0xc1,0x3c,0x6e,0x68,0x3e,0x0b,0x6b,0x3e,0x00,
0x9f,0x09,0x81,0xc3,0xc1,0x80,0xd8,0xfd,0x20,0x52,0x24,0x08,0x12,0x25,0x02,0x00,
0xc7,0x83,0x83,0x83,0x00,0x00,0x80,0xc6,0x18,0x24,0x34,0x38,0x4d,0x46,0x39,0x00,
0x8f,0x8f,0xcf,0x8f,0xdf,0xff,0xff,0xff,0x30,0x30,0x10,0x20,0x00,0x00,0x00,0x00,
0xf3,0xe3,0xc7,0xc7,0xc7,0xe7,0xf3,0xfb,0x04,0x08,0x18,0x10,0x18,0x08,0x04,0x00,
0xcf,0xe7,0xe3,0xf3,0xe3,0xe3,0xc7,0xef,0x10,0x08,0x0c,0x04,0x0c,0x08,0x10,0x00,
0xff,0xc9,0xe1,0x80,0xc0,0xc1,0xed,0xff,0x00,0x12,0x0c,0x3f,0x0c,0x12,0x00,0x00,
0xff,0xe3,0xe3,0x80,0xc0,0xe3,0xf3,0xff,0x00,0x0c,0x0c,0x3f,0x0c,0x0c,0x00,0x00,
0xff,0xff,0xff,0xff,0xc7,0xc7,0x87,0xcf,0x00,0x00,0x00,0x00,0x18,0x18,0x30,0x00,
0xff,0xff,0xff,0x81,0xc1,0xff,0xff,0xff,0x00,0x00,0x00,0x3e,0x00,0x00,0x00,0x00,
0xff,0xff,0xff,0xff,0xff,0x8f,0x8f,0xcf,0x00,0x00,0x00,0x00,0x00,0x30,0x30,0x00,
0xff,0xf9,0xf1,0xe3,0xc7,0x8f,0xdf,0xff,0x00,0x02,0x04,0x08,0x10,0x20,0x00,0x00,
0xc3,0x81,0x00,0x00,0x00,0x80,0xc1,0xe3,0x1c,0x26,0x63,0x63,0x63,0x32,0x1c,0x00,
0xe3,0xc3,0xe3,0xe3,0xe3,0xe3,0x80,0xc0,0x0c,0x1c,0x0c,0x0c,0x0c,0x0c,0x3f,0x00,
0x81,0x00,0x80,0xc0,0x81,0x03,0x00,0x80,0x3e,0x63,0x07,0x1e,0x3c,0x70,0x7f,0x00,
0x00,0x80,0xe1,0xc1,0xe0,0x00,0x80,0xc1,0x7f,0x06,0x0c,0x1e,0x03,0x63,0x3e,0x00,
0xe1,0xc1,0x81,0x01,0x00,0x80,0xf1,0xf9,0x0e,0x1e,0x36,0x66,0x7f,0x06,0x06,0x00,
0x01,0x01,0x01,0x80,0xf8,0x18,0x80,0xc1,0x7e,0x60,0x7e,0x03,0x03,0x63,0x3e,0x00,
0xc1,0x81,0x0f,0x01,0x00,0x00,0x80,0xc1,0x1e,0x30,0x60,0x7e,0x63,0x63,0x3e,0x00,
0x00,0x80,0xf0,0xe1,0xc3,0xc7,0xc7,0xe7,0x7f,0x03,0x06,0x0c,0x18,0x18,0x18,0x00,
0x83,0x01,0x01,0x81,0x00,0x00,0x80,0xc1,0x3c,0x62,0x72,0x3c,0x4f,0x43,0x3e,0x00,
0x81,0x00,0x00,0x80,0xc0,0xf0,0x81,0xc3,0x3e,0x63,0x63,0x3f,0x03,0x06,0x3c,0x00,
0xff,0xc7,0xc7,0xe7,0xc7,0xc7,0xe7,0xff,0x00,0x18,0x18,0x00,0x18,0x18,0x00,0x00,
0xff,0xc7,0xc7,0xe7,0xc7,0xc7,0x87,0xcf,0x00,0x18,0x18,0x00,0x18,0x18,0x30,0x00,
0xf1,0xe1,0xc3,0x87,0xc7,0xe3,0xf1,0xf9,0x06,0x0c,0x18,0x30,0x18,0x0c,0x06,0x00,
0xff,0xff,0x81,0xc1,0x81,0xc1,0xff,0xff,0x00,0x00,0x3e,0x00,0x3e,0x00,0x00,0x00,
0x8f,0xc7,0xe3,0xf1,0xe1,0xc3,0x87,0xcf,0x30,0x18,0x0c,0x06,0x0c,0x18,0x30,0x00,
0x81,0x00,0x00,0x80,0xe1,0xf3,0xe3,0xf3,0x3e,0x7f,0x63,0x06,0x0c,0x00,0x0c,0x00,
0x83,0x01,0x00,0x00,0x00,0x00,0x00,0x81,0x3c,0x42,0x99,0xa1,0xa1,0x99,0x42,0x3c,
0xc3,0x81,0x00,0x00,0x00,0x00,0x18,0x9c,0x1c,0x36,0x63,0x63,0x7f,0x63,0x63,0x00,
0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x81,0x7e,0x63,0x63,0x7e,0x63,0x63,0x7e,0x00,
0xc1,0x80,0x0c,0x1f,0x1f,0x88,0xc0,0xe1,0x1e,0x33,0x60,0x60,0x60,0x33,0x1e,0x00,
0x03,0x01,0x00,0x00,0x00,0x00,0x01,0x83,0x7c,0x66,0x63,0x63,0x63,0x66,0x7c,0x00,
0x00,0x00,0x1f,0x01,0x01,0x1f,0x00,0x80,0x7f,0x60,0x60,0x7e,0x60,0x60,0x7f,0x00,
0x00,0x00,0x1f,0x01,0x01,0x1f,0x1f,0x9f,0x7f,0x60,0x60,0x7e,0x60,0x60,0x60,0x00,
0xc0,0x80,0x0f,0x00,0x00,0x80,0xc0,0xe0,0x1f,0x30,0x60,0x6f,0x63,0x33,0x1f,0x00,
0x18,0x18,0x18,0x00,0x00,0x18,0x18,0x9c,0x63,0x63,0x63,0x7f,0x63,0x63,0x63,0x00,
0x80,0xc0,0xe3,0xe3,0xe3,0xe3,0x80,0xc0,0x3f,0x0c,0x0c,0x0c,0x0c,0x0c,0x3f,0x00,
0xf8,0xf8,0xf8,0xf8,0xf8,0x18,0x80,0xc1,0x03,0x03,0x03,0x03,0x03,0x63,0x3e,0x00,
0x18,0x10,0x01,0x03,0x03,0x01,0x10,0x98,0x63,0x66,0x6c,0x78,0x7c,0x6e,0x67,0x00,
0x8f,0x8f,0x8f,0x8f,0x8f,0x8f,0x80,0xc0,0x30,0x30,0x30,0x30,0x30,0x30,0x3f,0x00,
0x18,0x00,0x00,0x00,0x10,0x18,0x18,0x9c,0x63,0x77,0x7f,0x6b,0x63,0x63,0x63,0x00,
0x18,0x08,0x00,0x00,0x00,0x10,0x18,0x9c,0x63,0x73,0x7b,0x7f,0x6f,0x67,0x63,0x00,
0x81,0x00,0x00,0x00,0x00,0x00,0x80,0xc1,0x3e,0x63,0x63,0x63,0x63,0x63,0x3e,0x00,
0x01,0x00,0x00,0x00,0x00,0x01,0x1f,0x9f,0x7e,0x63,0x63,0x63,0x7e,0x60,0x60,0x00,
0x81,0x00,0x00,0x00,0x00,0x00,0x80,0xc2,0x3e,0x63,0x63,0x63,0x6f,0x66,0x3d,0x00,
0x01,0x00,0x00,0x00,0x00,0x01,0x10,0x98,0x7e,0x63,0x63,0x67,0x7c,0x6e,0x67,0x00,
0x83,0x01,0x01,0x81,0xc0,0x00,0x80,0xc1,0x3c,0x66,0x60,0x3e,0x03,0x63,0x3e,0x00,
0x80,0xc0,0xe3,0xe3,0xe3,0xe3,0xe3,0xf3,0x3f,0x0c,0x0c,0x0c,0x0c,0x0c,0x0c,0x00,
0x18,0x18,0x18,0x18,0x18,0x18,0x80,0xc1,0x63,0x63,0x63,0x63,0x63,0x63,0x3e,0x00,
0x18,0x18,0x18,0x00,0x80,0xc1,0xe3,0xf7,0x63,0x63,0x63,0x77,0x3e,0x1c,0x08,0x00,
0x18,0x18,0x18,0x00,0x00,0x00,0x08,0x9c,0x63,0x63,0x63,0x6b,0x7f,0x77,0x63,0x00,
0x18,0x00,0x80,0xc1,0x81,0x00,0x08,0x9c,0x63,0x77,0x3e,0x1c,0x3e,0x77,0x63,0x00,
0x11,0x11,0x11,0x81,0xc3,0xc7,0xc7,0xe7,0x66,0x66,0x66,0x3c,0x18,0x18,0x18,0x00,
0x00,0x80,0xe0,0xc1,0x83,0x07,0x00,0x80,0x7f,0x07,0x0e,0x1c,0x38,0x70,0x7f,0x00,
0xc3,0xc3,0xc7,0xc7,0xc7,0xc7,0xc3,0xe3,0x1c,0x18,0x18,0x18,0x18,0x18,0x1c,0x00,
0xff,0x9f,0xcf,0xe7,0xf3,0xf9,0xfd,0xff,0x00,0x20,0x10,0x08,0x04,0x02,0x00,0x00,
0xc3,0xe3,0xe3,0xe3,0xe3,0xe3,0xc3,0xe3,0x1c,0x0c,0x0c,0x0c,0x0c,0x0c,0x1c,0x00,
0xe7,0xc3,0x81,0x08,0x9c,0xff,0xff,0xff,0x08,0x1c,0x36,0x63,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x80,0x7f,0x7f,0x7f,0x7f,0x7f,0x7f,0x7f,0x00
};


/**********************************************************************/
/* tables trigonometriques 14 bits */

static int tsn[320]={
    0,402,804,1205,1606,2006,2404,2801,
    3196,3590,3981,4370,4756,5139,5520,5897,
    6270,6639,7005,7366,7723,8076,8423,8765,
    9102,9434,9760,10080,10394,10702,11003,11297,
    11585,11866,12140,12406,12665,12916,13160,13395,
    13623,13842,14053,14256,14449,14635,14811,14978,
    15137,15286,15426,15557,15679,15791,15893,15986,
    16069,16143,16207,16261,16305,16340,16364,16379,
    16384,16379,16364,16340,16305,16261,16207,16143,
    16069,15986,15893,15791,15679,15557,15426,15286,
    15137,14978,14811,14635,14449,14256,14053,13842,
    13623,13395,13160,12916,12665,12406,12140,11866,
    11585,11297,11003,10702,10394,10080,9760,9434,
    9102,8765,8423,8076,7723,7366,7005,6639,
    6270,5897,5520,5139,4756,4370,3981,3590,
    3196,2801,2404,2006,1606,1205,804,402,
    0,-402,-804,-1205,-1606,-2006,-2404,-2801,
    -3196,-3590,-3981,-4370,-4756,-5139,-5520,-5897,
    -6270,-6639,-7005,-7366,-7723,-8076,-8423,-8765,
    -9102,-9434,-9760,-10080,-10394,-10702,-11003,-11297,
    -11585,-11866,-12140,-12406,-12665,-12916,-13160,-13395,
    -13623,-13842,-14053,-14256,-14449,-14635,-14811,-14978,
    -15137,-15286,-15426,-15557,-15679,-15791,-15893,-15986,
    -16069,-16143,-16207,-16261,-16305,-16340,-16364,-16379,
    -16384,-16379,-16364,-16340,-16305,-16261,-16207,-16143,
    -16069,-15986,-15893,-15791,-15679,-15557,-15426,-15286,
    -15137,-14978,-14811,-14635,-14449,-14256,-14053,-13842,
    -13623,-13395,-13160,-12916,-12665,-12406,-12140,-11866,
    -11585,-11297,-11003,-10702,-10394,-10080,-9760,-9434,
    -9102,-8765,-8423,-8076,-7723,-7366,-7005,-6639,
    -6270,-5897,-5520,-5139,-4756,-4370,-3981,-3590,
    -3196,-2801,-2404,-2006,-1606,-1205,-804,-402,
    0,402,804,1205,1606,2006,2404,2801,
    3196,3590,3981,4370,4756,5139,5520,5897,
    6270,6639,7005,7366,7723,8076,8423,8765,
    9102,9434,9760,10080,10394,10702,11003,11297,
    11585,11866,12140,12406,12665,12916,13160,13395,
    13623,13842,14053,14256,14449,14635,14811,14978,
    15137,15286,15426,15557,15679,15791,15893,15986,
    16069,16143,16207,16261,16305,16340,16364,16379
};
static int ttg[65]={
    -16384,-15599,-14850,-14133,-13446,-12786,-12151,-11539,
    -10947,-10375,-9820,-9281,-8757,-8247,-7749,-7263,
    -6786,-6320,-5862,-5413,-4970,-4534,-4104,-3679,
    -3259,-2843,-2430,-2021,-1614,-1209,-805,-402,
    0,402,805,1209,1614,2021,2430,2843,
    3259,3679,4104,4534,4970,5413,5862,6320,
    6786,7263,7749,8247,8757,9281,9820,10375,
    10947,11539,12151,12786,13446,14133,14850,15599,
    16384
};
int *sn=tsn;
int *cs=tsn+64;
int *tg=ttg+32;




/* variables pour Direct X */

/* Direct Draw */

#define DD						0
#define DESKTOP					1

static	int						modmode=DD;

static	LPDIRECTDRAWPALETTE		lpPalette=NULL;
static	LPDIRECTDRAWSURFACE		lpFrontBuffer=NULL;
static	LPDIRECTDRAWSURFACE		lpInterBuffer;
static	LPDIRECTDRAWSURFACE		lpBackBuffer;
static	LPDIRECTDRAW			lpDD;
	
static	unsigned char			*DirectXBuffer=NULL;

static	long					video_offset=0;
static	int						y_offset=0;

/* Direct Sound */


typedef struct
{
    BYTE *pbWaveData;               // pointer into wave resource (for restore)
    DWORD cbWaveSize;               // size of wave data (for restore)
    int iAlloc;                     // number of buffers.
    int iCurrent;                   // current buffer
    IDirectSoundBuffer* Buffers[1]; // list of buffers

} SNDOBJ, *HSNDOBJ;


IDirectSoundBuffer *SndObjGetFreeBuffer(HSNDOBJ hSO);

static	int						bSoundEnabled = TRUE;

static	LPDIRECTSOUND           lpDS;

static	HSNDOBJ					hsoSample[MAXSAMPLE];


/* For compatibility with WING modules */
long      Orientation = -1;     /* assume up-bottom DIBs*/



void alert(char *titre,long n)
{
	char s[250];

	sprintf(s,"(%lx)",n);
    MessageBox(GetFocus(),s,titre,0);
}




/***********************************************************/
/* Direct Sound */

char * dserrorstring(long n)
{
	if (n==DS_OK)
		return "No error";
	if (n==DSERR_ALLOCATED)
		return "The call failed because resources (such as a priority level) were already being used by another caller.";
	if (n==DSERR_CONTROLUNAVAIL)
		return "The control (vol,pan,etc.) requested by the caller is not available.";
	if (n==DSERR_INVALIDPARAM)
		return "An invalid parameter was passed to the returning function";
	if (n==DSERR_INVALIDCALL)
		return "This call is not valid for the current state of this object";
	if (n==DSERR_GENERIC)
		return "The caller does not have the priority level required for the function to succeed.";
	if (n==DSERR_OUTOFMEMORY)
		return "The DSound subsystem couldn't allocate sufficient memory to complete the caller's request.";
	if (n==DSERR_BADFORMAT)
		return "The specified WAVE format is not supported";
	if (n==DSERR_UNSUPPORTED)
		return "The function called is not supported at this time";
	if (n==DSERR_NODRIVER)
		return "No sound driver is available for use";
	if (n==DSERR_ALREADYINITIALIZED)
		return "This object is already initialized";
	if (n==DSERR_NOAGGREGATION)
		return "This object does not support aggregation";
	if (n==DSERR_BUFFERLOST)
		return "The buffer memory has been lost, and must be Restored.";
	if (n==DSERR_OTHERAPPHASPRIO)
		return "Another app has a higher priority level, preventing this call from succeeding.";
	if (n==DSERR_UNINITIALIZED)
		return "The Initialize() member on the Direct Sound Object has not been called or called successfully before calls to other members.";
 
	return "Unknown error";
};


void dsalert(char *titre,long n)
{
	char s[250];

	if( n != DS_OK )
	{
		sprintf(s,"%s (%lx)",dserrorstring(n),n);
	    MessageBox(GetFocus(),s,titre,0);
	}
}





BOOL DSFillSoundBuffer(IDirectSoundBuffer *pDSB, BYTE *pbWaveData, DWORD cbWaveSize)
{
    if (pDSB && pbWaveData && cbWaveSize)
    {
        LPVOID pMem1, pMem2;
        DWORD dwSize1, dwSize2;

        if (SUCCEEDED(IDirectSoundBuffer_Lock(pDSB, 0, cbWaveSize,
            &pMem1, &dwSize1, &pMem2, &dwSize2, 0)))
        {
            CopyMemory(pMem1, pbWaveData, dwSize1);

            if ( 0 != dwSize2 )
                CopyMemory(pMem2, pbWaveData+dwSize1, dwSize2);

            IDirectSoundBuffer_Unlock(pDSB, pMem1, dwSize1, pMem2, dwSize2);
            return TRUE;
        }
    }

    return FALSE;
}


void SndObjDestroy(SNDOBJ *pSO)
{
    if (pSO)
    {
        int i;

        for (i=0; i<pSO->iAlloc; i++)
        {
            if (pSO->Buffers[i])
            {
                IDirectSoundBuffer_Release(pSO->Buffers[i]);
                pSO->Buffers[i] = NULL;
            }
        }
        LocalFree((HANDLE)pSO);
    }
}


BOOL SndObjPlay(SNDOBJ *pSO, DWORD dwPlayFlags)
{
    BOOL result = FALSE;

    if (pSO == NULL)
        return FALSE;

    if ((!(dwPlayFlags & DSBPLAY_LOOPING) || (pSO->iAlloc == 1)))
    {
        IDirectSoundBuffer *pDSB = SndObjGetFreeBuffer(pSO);
        if (pDSB != NULL) {
            result = SUCCEEDED(IDirectSoundBuffer_Play(pDSB, 0, 0, dwPlayFlags));
        }
    }

    return result;
}


BOOL SndObjStop(SNDOBJ *pSO)
{
    int i;

    if (pSO == NULL)
        return FALSE;

    for (i=0; i<pSO->iAlloc; i++)
    {
        IDirectSoundBuffer_Stop(pSO->Buffers[i]);
        IDirectSoundBuffer_SetCurrentPosition(pSO->Buffers[i], 0);
    }

    return TRUE;
}


IDirectSoundBuffer *SndObjGetFreeBuffer(SNDOBJ *pSO)
{
    IDirectSoundBuffer *pDSB;

    if (pSO == NULL)
        return NULL;

    if (pDSB = pSO->Buffers[pSO->iCurrent])
    {
        HRESULT hres;
        DWORD dwStatus;

        hres = IDirectSoundBuffer_GetStatus(pDSB, &dwStatus);

        if (FAILED(hres))
            dwStatus = 0;

        if ((dwStatus & DSBSTATUS_PLAYING) == DSBSTATUS_PLAYING)
        {
            if (pSO->iAlloc > 1)
            {
                if (++pSO->iCurrent >= pSO->iAlloc)
                    pSO->iCurrent = 0;

                pDSB = pSO->Buffers[pSO->iCurrent];
                hres = IDirectSoundBuffer_GetStatus(pDSB, &dwStatus);

                if (SUCCEEDED(hres) && (dwStatus & DSBSTATUS_PLAYING) == DSBSTATUS_PLAYING)
                {
                    IDirectSoundBuffer_Stop(pDSB);
                    IDirectSoundBuffer_SetCurrentPosition(pDSB, 0);
                }
            }
            else
            {
                pDSB = NULL;
            }
        }

        if (pDSB && (dwStatus & DSBSTATUS_BUFFERLOST))
        {
            if (FAILED(IDirectSoundBuffer_Restore(pDSB)) ||
                !DSFillSoundBuffer(pDSB, pSO->pbWaveData, pSO->cbWaveSize))
            {
                pDSB = NULL;
            }
        }
    }

    return pDSB;
}




IDirectSoundBuffer *DSLoadSoundBufferFromMemory(IDirectSound *pDS, LPWAVEFORMATEX pWaveHeader, BYTE *pbData, UINT cbData)
{
    IDirectSoundBuffer *pDSB = NULL;
    DSBUFFERDESC dsBD = {0};
    BYTE *pbWaveData;
	long n;

	pbWaveData=pbData;
	dsBD.lpwfxFormat=pWaveHeader;
	dsBD.dwBufferBytes=cbData;
    
    {
        dsBD.dwSize = sizeof(dsBD);
        dsBD.dwFlags = DSBCAPS_STATIC | DSBCAPS_CTRLDEFAULT;


        if (SUCCEEDED(IDirectSound_CreateSoundBuffer(pDS, &dsBD, &pDSB, NULL)))
        {

            if (!DSFillSoundBuffer(pDSB, pbWaveData, dsBD.dwBufferBytes))
            {
                IDirectSoundBuffer_Release(pDSB);
                pDSB = NULL;
            }
        }
        else
        {
			dsalert("CreateSoundBuffer",n);
            pDSB = NULL;
        }
    }

    return pDSB;
}


SNDOBJ *SndObjCreateFromMemory(IDirectSound *pDS, int iConcurrent, LPWAVEFORMATEX pWaveHeader, BYTE *pbData, UINT cbData)
{
    SNDOBJ *pSO = NULL;
 
    {
        if (iConcurrent < 1)
            iConcurrent = 1;

        if ((pSO = (SNDOBJ *)LocalAlloc(LPTR, sizeof(SNDOBJ) +
            (iConcurrent-1) * sizeof(IDirectSoundBuffer *))) != NULL)
        {
            int i;

            pSO->iAlloc = iConcurrent;
            pSO->pbWaveData = pbData;
            pSO->cbWaveSize = cbData;
            pSO->Buffers[0] = DSLoadSoundBufferFromMemory(pDS, pWaveHeader, pbData, cbData);

            for (i=1; i<pSO->iAlloc; i++)
            {
                if (FAILED(IDirectSound_DuplicateSoundBuffer(pDS,
                    pSO->Buffers[0], &pSO->Buffers[i])))
                {
                    pSO->Buffers[i] = DSLoadSoundBufferFromMemory(pDS, pWaveHeader, pbData, cbData);
                    if (!pSO->Buffers[i]) 
					{
                        SndObjDestroy(pSO);
                        pSO = NULL;
                        break;
                    }
                }
            }
        }
    }

    return pSO;
}





void InitializeSound( void )
{
    bSoundEnabled = FALSE;
    if (SUCCEEDED(DirectSoundCreate(NULL, &lpDS, NULL)))
    {
        if (SUCCEEDED(lpDS->lpVtbl->SetCooperativeLevel(lpDS, window,
            DSSCL_NORMAL)))
        {

			bSoundEnabled = TRUE;
        }
        else
        {
            lpDS->lpVtbl->Release(lpDS);
            lpDS = NULL;
        }
    }
}

void DestroySound( void )
{
	int i;

    bSoundEnabled = FALSE;
    if (lpDS)
    {
	
		for(i=0;i<MAXSAMPLE;i++)
		if (splfrq[i])
		{	                         
			SndObjDestroy(hsoSample[i]);
			hsoSample[i] = NULL;
		}

        
        lpDS->lpVtbl->Release(lpDS);
        lpDS = NULL;
    }
}









/***********************************************************/
/* Direct Draw */

char * dderrorstring(long n)
{	
	if (n==DD_OK)
		return "No error";
	if (n==DDERR_ALREADYINITIALIZED)
		return "This object is already initialized";
	if (n==DDERR_CANNOTATTACHSURFACE)
		return "This surface can not be attached to the requested surface.";
	if (n==DDERR_CANNOTDETACHSURFACE)
		return "This surface can not be detached from the requested surface.";
	if (n==DDERR_CURRENTLYNOTAVAIL)
		return "Support is currently not available.";
	if (n==DDERR_EXCEPTION)
		return "An exception was encountered while performing the requested operation";
	if (n==DDERR_GENERIC)
		return "Generic failure.";
	if (n==DDERR_HEIGHTALIGN)
		return "Height of rectangle provided is not a multiple of reqd alignment";
	if (n==DDERR_INCOMPATIBLEPRIMARY)
		return "Unable to match primary surface creation request with existing primary surface.";
	if (n==DDERR_INVALIDCAPS)
		return "One or more of the caps bits passed to the callback are incorrect.";
	if (n==DDERR_INVALIDCLIPLIST)
		return "DirectDraw does not support provided Cliplist.";
	if (n==DDERR_INVALIDMODE)
		return "DirectDraw does not support the requested mode";
	if (n==DDERR_INVALIDOBJECT)
		return "DirectDraw received a pointer that was an invalid DIRECTDRAW object.";
	if (n==DDERR_INVALIDPARAMS)
		return "One or more of the parameters passed to the callback function are incorrect.";
	if (n==DDERR_INVALIDPIXELFORMAT)
		return "pixel format was invalid as specified";
	if (n==DDERR_INVALIDRECT)
		return "Rectangle provided was invalid.";
	if (n==DDERR_LOCKEDSURFACES)
		return "Operation could not be carried out because one or more surfaces are locked";
	if (n==DDERR_NO3D)
		return "There is no 3D present.";
	if (n==DDERR_NOALPHAHW)
		return "Operation could not be carried out because there is no alpha accleration hardware present or available.";
	if (n==DDERR_NOCLIPLIST)
		return "no clip list available";
	if (n==DDERR_NOCOLORCONVHW)
		return "Operation could not be carried out because there is no color conversion hardware present or available.";
	if (n==DDERR_NOCOOPERATIVELEVELSET)
		return "Create function called without DirectDraw object method SetCooperativeLevel being called.";
	if (n==DDERR_NOCOLORKEY)
		return "Surface doesn't currently have a color key";
	if (n==DDERR_NOCOLORKEYHW)
		return "Operation could not be carried out because there is no hardware support of the dest color key.";
	if (n==DDERR_NODIRECTDRAWSUPPORT)
		return "No DirectDraw support possible with current display driver";
	if (n==DDERR_NOEXCLUSIVEMODE)
		return "Operation requires the application to have exclusive mode but the application does not have exclusive mode.";
	if (n==DDERR_NOFLIPHW)
		return "Flipping visible surfaces is not supported.";
	if (n==DDERR_NOGDI)
		return "There is no GDI present.";
	if (n==DDERR_NOMIRRORHW)
		return "Operation could not be carried out because there is no hardware present or available.";
	if (n==DDERR_NOTFOUND)
		return "Requested item was not found";
	if (n==DDERR_NOOVERLAYHW)
		return "Operation could not be carried out because there is no overlay hardware present or available.";
	if (n==DDERR_NORASTEROPHW)
		return "Operation could not be carried out because there is no appropriate raster op hardware present or available.";
	if (n==DDERR_NOROTATIONHW)
		return "Operation could not be carried out because there is no rotation hardware present or available.";
	if (n==DDERR_NOSTRETCHHW)
		return "Operation could not be carried out because there is no hardware support for stretching";
	if (n==DDERR_NOT4BITCOLOR)
		return "DirectDrawSurface is not in 4 bit color palette and the requested operation requires 4 bit color palette.";
	if (n==DDERR_NOT4BITCOLORINDEX)
		return "DirectDrawSurface is not in 4 bit color index palette and the requested operation requires 4 bit color index palette.";
	if (n==DDERR_NOT8BITCOLOR)
		return "DirectDraw Surface is not in 8 bit color mode and the requested operation requires 8 bit color.";
	if (n==DDERR_NOTEXTUREHW)
		return "Operation could not be carried out because there is no texture mapping hardware present or available.";
	if (n==DDERR_NOVSYNCHW)
		return "Operation could not be carried out because there is no hardware support for vertical blank synchronized operations.";
	if (n==DDERR_NOZBUFFERHW)
		return "Operation could not be carried out because there is no hardware support for zbuffer blting.";
	if (n==DDERR_NOZOVERLAYHW)
		return "Overlay surfaces could not be z layered based on their BltOrder because the hardware does not support z layering of overlays.";
	if (n==DDERR_OUTOFCAPS)
		return "The hardware needed for the requested operation has already been allocated.";
	if (n==DDERR_OUTOFMEMORY)
		return "DirectDraw does not have enough memory to perform the operation.";
	if (n==DDERR_OUTOFVIDEOMEMORY)
		return "DirectDraw does not have enough memory to perform the operation.";
	if (n==DDERR_OVERLAYCANTCLIP)
		return "hardware does not support clipped overlays";
	if (n==DDERR_OVERLAYCOLORKEYONLYONEACTIVE)
		return "Can only have ony color key active at one time for overlays";
	if (n==DDERR_PALETTEBUSY)
		return "Access to this palette is being refused because the palette is already locked by another thread.";
	if (n==DDERR_COLORKEYNOTSET)
		return "No src color key specified for this operation.";
	if (n==DDERR_SURFACEALREADYATTACHED)
		return "This surface is already attached to the surface it is being attached to.";
	if (n==DDERR_SURFACEALREADYDEPENDENT)
		return "This surface is already a dependency of the surface it is being made a dependency of.";
	if (n==DDERR_SURFACEBUSY)
		return "Access to this surface is being refused because the surface is already locked by another thread.";
	if (n==DDERR_SURFACEISOBSCURED)
		return "Access to Surface refused because Surface is obscured.";
	if (n==DDERR_SURFACELOST)
		return "Access to this surface is being refused because the surface is gone. The DIRECTDRAWSURFACE object representing this surface should have Restore called on it.";
	if (n==DDERR_SURFACENOTATTACHED)
		return "The requested surface is not attached.";
	if (n==DDERR_TOOBIGHEIGHT)
		return "Height requested by DirectDraw is too large.";
	if (n==DDERR_TOOBIGSIZE)
		return "Size requested by DirectDraw is too large --  The individual height and width are OK.";
	if (n==DDERR_TOOBIGWIDTH)
		return "Width requested by DirectDraw is too large.";
	if (n==DDERR_UNSUPPORTED)
		return "Action not supported.";
	if (n==DDERR_UNSUPPORTEDFORMAT)
		return "FOURCC format requested is unsupported by DirectDraw";
	if (n==DDERR_UNSUPPORTEDMASK)
		return "Bitmask in the pixel format requested is unsupported by DirectDraw";
	if (n==DDERR_VERTICALBLANKINPROGRESS)
		return "vertical blank is in progress";
	if (n==DDERR_WASSTILLDRAWING)
		return "Informs DirectDraw that the previous Blt which is transfering information to or from this Surface is incomplete.";
	if (n==DDERR_XALIGN)
		return "Rectangle provided was not horizontally aligned on reqd. boundary";
	if (n==DDERR_INVALIDDIRECTDRAWGUID)
		return "The GUID passed to DirectDrawCreate is not a valid DirectDraw driver identifier.";
	if (n==DDERR_DIRECTDRAWALREADYCREATED)
		return "A DirectDraw object representing this driver has already been created for this process.";
	if (n==DDERR_NODIRECTDRAWHW)
		return "A hardware only DirectDraw object creation was attempted but the driver did not support any hardware.";
	if (n==DDERR_PRIMARYSURFACEALREADYEXISTS)
		return "this process already has created a primary surface";
	if (n==DDERR_NOEMULATION)
		return "software emulation not available.";
	if (n==DDERR_REGIONTOOSMALL)
		return "region passed to Clipper::GetClipList is too small.";
	if (n==DDERR_CLIPPERISUSINGHWND)
		return "an attempt was made to set a clip list for a clipper objec that is already monitoring an hwnd.";
	if (n==DDERR_NOCLIPPERATTACHED)
		return "No clipper object attached to surface object";
	if (n==DDERR_NOHWND)
		return "Clipper notification requires an HWND or no HWND has previously been set as the CooperativeLevel HWND.";
	if (n==DDERR_HWNDSUBCLASSED)
		return "HWND used by DirectDraw CooperativeLevel has been subclassed, this prevents DirectDraw from restoring state.";
	if (n==DDERR_HWNDALREADYSET)
		return "The CooperativeLevel HWND has already been set. It can not be reset while the process has surfaces or palettes created.";
	if (n==DDERR_NOPALETTEATTACHED)
		return "No palette object attached to this surface.";
	if (n==DDERR_NOPALETTEHW)
		return "No hardware support for 16 or 256 color palettes.";
	if (n==DDERR_BLTFASTCANTCLIP)
		return "If a clipper object is attached to the source surface passed into a BltFast call.";
	if (n==DDERR_NOBLTHW)
		return "No blter.";
	if (n==DDERR_NODDROPSHW)
		return "No DirectDraw ROP hardware.";
	if (n==DDERR_OVERLAYNOTVISIBLE)
		return "returned when GetOverlayPosition is called on a hidden overlay";
	if (n==DDERR_NOOVERLAYDEST)
		return "returned when GetOverlayPosition is called on a overlay that UpdateOverlay has never been called on to establish a destionation.";
	if (n==DDERR_INVALIDPOSITION)
		return "returned when the position of the overlay on the destionation is no longer legal for that destionation.";
	if (n==DDERR_NOTAOVERLAYSURFACE)
		return "returned when an overlay member is called for a non-overlay surface";
	if (n==DDERR_EXCLUSIVEMODEALREADYSET)
		return "An attempt was made to set the cooperative level when it was already set to exclusive.";
	if (n==DDERR_NOTFLIPPABLE)
		return "An attempt has been made to flip a surface that is not flippable.";
	if (n==DDERR_CANTDUPLICATE)
		return "Can't duplicate primary & 3D surfaces, or surfaces that are implicitly created.";
	if (n==DDERR_NOTLOCKED)
		return "Surface was not locked.  An attempt to unlock a surface that was not locked at all, or by this process, has been attempted.";
	if (n==DDERR_CANTCREATEDC)
		return "Windows can not create any more DCs";
	if (n==DDERR_NODC)
		return "No DC was ever created for this surface.";
	if (n==DDERR_NODC)
		return "This surface can not be restored because it was created in a different mode.";
	if (n==DDERR_IMPLICITLYCREATED)
		return "This surface can not be restored because it is an implicitly created surface.";
	if (n==DDERR_NOTPALETTIZED)
		return "The surface being used is not a palette-based surface";
	if (n==DDERR_UNSUPPORTEDMODE)
		return "The display is currently in an unsupported mode";

	return "Unknown error";
};



void ddalert(char *titre,long n)
{
	char s[250];

	if( n != DD_OK )
	{
		sprintf(s,"%s (%lx)",dderrorstring(n),n);
	    MessageBox(GetFocus(),s,titre,0);
	}
}


void alerte(char *message,char *titre,long n)
{
	char s[250];

	sprintf(s,"%s (%lx)",message,n);
	
    MessageBox(GetFocus(),s,titre,0);
}




/*******************************************************************************/
/* initialisation du systeme */


unsigned char *lastworkscreen=0;

unsigned char *lockworkscreen(void)
{
    DDSURFACEDESC   ddsd;
    HRESULT			ddrval;


	if (modmode!=DD) return NULL;

	if (lpBackBuffer==NULL) return NULL;

	if (lastworkscreen) return lastworkscreen;

	memset( &ddsd, 0, sizeof( ddsd ) );
    ddsd.dwSize = sizeof( ddsd );

	ddrval=lpBackBuffer->lpVtbl->Lock(lpBackBuffer,NULL,&ddsd,DDLOCK_WAIT,NULL);
	if (ddrval==DDERR_SURFACELOST)
	{
		if (!RestoreSurfaces()) return 0;
	}
	else
		ddalert("Lock work screen !",ddrval);

	if (ddsd.lPitch) BPL=ddsd.lPitch;
	if (video_offset_not_null) video_offset=40*BPL*HEIGHT*RATIO/400;
	//in internal buffer number of bytes per line
	//is not always equal to RATIO*WIDTH, and we
	//can know it only here at execution time

	DirectXBuffer=(unsigned char *)ddsd.lpSurface;

	if (!DirectXBuffer) return 0;

	lastworkscreen=DirectXBuffer+video_offset;

	return DirectXBuffer+video_offset;
}



void releaseworkscreen(void)
{
    DDSURFACEDESC   ddsd;
    HRESULT			ddrval;



	if (modmode!=DD) return;

	if (lpBackBuffer==NULL) return;

	if (DirectXBuffer)
	{
		memset( &ddsd, 0, sizeof( ddsd ) );
		ddsd.dwSize = sizeof( ddsd );

		ddrval=lpBackBuffer->lpVtbl->Unlock(lpBackBuffer,DirectXBuffer);
		ddalert("Unlock work screen",ddrval);
		DirectXBuffer=NULL;
	}
}







static void initwindow(void)
{
	DDCAPS          ddcaps;
    HRESULT         ddrval;
    DDSURFACEDESC   ddsd;
    DDSCAPS         ddscaps;
    PALETTEENTRY    ape[256];

    unsigned long i;
    unsigned char huge *p;     
    
	
	InitializeSound();                        


    // Create surfaces
    memset( &ddsd, 0, sizeof( ddsd ) );
    ddsd.dwSize = sizeof( ddsd );

    ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE |
                          DDSCAPS_FLIP |
                          DDSCAPS_COMPLEX;
    ddsd.dwBackBufferCount = 1;
    ddrval = lpDD->lpVtbl->CreateSurface( lpDD, &ddsd, &lpFrontBuffer, NULL );
    if( ddrval != DD_OK )
        ddalert("CreateSurface",ddrval);

    // get a pointer to the back buffer
    ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
    ddrval = lpFrontBuffer->lpVtbl->GetAttachedSurface(
                lpFrontBuffer,
                &ddscaps,
                &lpBackBuffer );
    if( ddrval != DD_OK )
        ddalert("CreateSurface",ddrval);


    for (i=0; i<16; i++)
    {
        ape[i].peRed   = (BYTE)(((oldpal[i] >> 8) & 0x0F) * 255 / 15);
        ape[i].peGreen = (BYTE)(((oldpal[i] >> 4) & 0x0F) * 255 / 15);
        ape[i].peBlue  = (BYTE)(((oldpal[i] >> 0) & 0x0F) * 255 / 15);
        ape[i].peFlags = (BYTE)0;
    }

	ddalert("CreatePalette",lpDD->lpVtbl->CreatePalette(lpDD,DDPCAPS_8BIT, ape, &lpPalette, NULL));
	ddalert("SetPalette",lpFrontBuffer->lpVtbl->SetPalette( lpFrontBuffer, lpPalette));
}



int initsystem(void)
{

    setpalette(oldpal);
    cls();
    swap();
    cls();


    SetCursor(visible_cursor);

    setcolor(15);

    return 1;
}


void killcursor(void)
{
    SetCursor(visible_cursor);
}



static void killsystem2(void)
{
    MSG msg;
                   
	int i;
	char filename[80];
	char command[80];

    DestroySound();                                                

    if( lpFrontBuffer != NULL )
        lpFrontBuffer->lpVtbl->Release( lpFrontBuffer );

    if( lpPalette != NULL )
        lpPalette->lpVtbl->Release( lpPalette );

	if( lpDD != NULL )
		lpDD->lpVtbl->Release( lpDD );



    PostQuitMessage(0);

    killcursor();

    while(GetMessage(&msg, NULL, 0, 0))
    {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    freebloc(&cursorbak);
    freebloc(&cursorbob);

/*  exit(0);*/
    longjmp(jmpb,1);
}
    



void killsystem(void)
{
    MSG msg;
    
    DestroyWindow(window);    
 
    while(GetMessage(&msg, NULL, 0, 0))
    {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }
}



          
long FAR PASCAL WndProc (HWND hWnd, UINT Message, WPARAM wParam, LPARAM lParam)
{
	unsigned char *ptr;
	HDC hdc;

    BOOL f;            


    switch (Message)
    {
        case WM_DESTROY:
            killsystem2();
            break;

        case WM_ACTIVATEAPP:
			if (wParam)
				modmode=DD;
			else
			{
				modmode=DESKTOP;
				lastworkscreen=0;
			}
			break;
 
        case WM_PAINT:
			hdc = BeginPaint( hWnd, &ps );
			EndPaint( hWnd, &ps );
			return 1;
            break;



        case WM_SETCURSOR:


            if (realmsemode>0)
                SetCursor(visible_cursor);
            else 
                SetCursor(invisible_cursor);
            break;        
            

        default:
			return (DefWindowProc(hWnd, Message, wParam, lParam));
    }


    return (NULL);
}



static void initcursor(HANDLE hInstance)
{
    invisible_cursor=LoadCursor(hInstance,MAKEINTRESOURCE(101));  
    visible_cursor=LoadCursor(NULL, IDC_ARROW);
}


static int InitApplication(HANDLE hInstance)
{
    WNDCLASS  wc;

    wc.style = CS_BYTEALIGNCLIENT;
    wc.lpfnWndProc = WndProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hInstance;
	wc.hIcon = LoadIcon( hInstance, IDI_APPLICATION);
    wc.hCursor= NULL;
    wc.hbrBackground = GetStockObject( BLACK_BRUSH );
    wc.lpszMenuName =  NULL;
    wc.lpszClassName = NOM_CLASSE;

					     
	RegisterClass(&wc);

	return 1;
}

                 


static int InitInstance(HANDLE hInstance, int nCmdShow)
{
    HWND    hWnd;

    HDC Screen;
    int PaletteDevice;
    int nBitsPixel;         
    int nPlanes;         
    int nColorDepth;

	int cbmode;

	HRESULT ddrval;

    /* Detect non-palettized device */
    Screen = GetDC(0);

    if (Screen)
    {
        PaletteDevice = GetDeviceCaps(Screen, RASTERCAPS) & RC_PALETTE;
        nBitsPixel = GetDeviceCaps(Screen, BITSPIXEL);         
        nPlanes = GetDeviceCaps(Screen, PLANES);         
        nColorDepth = nBitsPixel * nPlanes;
        ReleaseDC(0, Screen);
/*
		if (nBitsPixel!=8)
		{
			MessageBox(GetFocus(),"Please, switch to 256 colors mode.","Wrong Display Mode",0);
			return FALSE;
		}
*/
	}




    
    hInst = hInstance;

    cbmode=WS_SYSMENU;

#ifdef NOCLOSEBOX
    cbmode=0;
#endif


    hWnd = CreateWindow(
        NOM_CLASSE,
        NOM_FENETRE,
		WS_VISIBLE | WS_POPUP | cbmode,
        0,
        0,
        WIDTH*RATIO,
        HEIGHT*RATIO,
        NULL,
        NULL,
        hInstance,
        NULL
        );

    if (!hWnd)
        return FALSE;

    window=hWnd;

	UpdateWindow(hWnd);

#ifdef DIRECTX_EMULATION
    ddrval = DirectDrawCreate( (LPVOID) DDCREATE_EMULATIONONLY, &lpDD, NULL );
#else
    ddrval = DirectDrawCreate( NULL, &lpDD, NULL );
#endif
	
    if( ddrval != DD_OK )
        ddalert("DirectDrawCreate",ddrval);

    ddrval = lpDD->lpVtbl->SetCooperativeLevel( lpDD, window,
                            DDSCL_ALLOWMODEX | DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN );

	
    if( ddrval != DD_OK )
        ddalert("SetCooperativeLevel",ddrval);

	if (WIDTH*RATIO==800)
		ddrval = lpDD->lpVtbl->SetDisplayMode( lpDD, 800, 600, 8 );
	else
	{
		if ((HEIGHT*RATIO==400)  || (HEIGHT*RATIO==200))
		{
			video_offset=40*BPL*HEIGHT*RATIO/400; //will be adjusted in lockworkscreen()
			video_offset_not_null=1;
			y_offset=40*HEIGHT*RATIO/400;
			ddrval = lpDD->lpVtbl->SetDisplayMode( lpDD, WIDTH*RATIO, 480*HEIGHT*RATIO/400, 8 );
		}
		else
			ddrval = lpDD->lpVtbl->SetDisplayMode( lpDD, WIDTH*RATIO, HEIGHT*RATIO, 8 );
	}


    if( ddrval != DD_OK )
        ddalert("SetDisplayMode",ddrval);

    return (TRUE);
}


static int get_args(char* argv[],char *commandline)
{
    int n=1;
    char *p=commandline;

    argv[0]="";

    while (*p)
    {
        argv[n]=p;
                n++;
        while ((*p)&&(*p!=' '))
            p++;
        if (*p==' ')
        {
            *p=0;
                        p++;
            while ((*p)&&(*p==' '))
                p++;
        }
    }

    return n;
}



#ifdef USER_FREE_FUNCTION
void free_all(void);
#endif

int main(int argc, char *argv[]);
               
static char commandline[256];

#ifdef BORLANDC
#pragma argsused
#endif
int PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
    int result=0;
    int     argc;
    char*   argv[20];

    if (!setjmp(jmpb))
    {             
#ifdef ONLY_ONCE
        if (hPrevInstance)
        {
			BringWindowToTop(FindWindow(NOM_CLASSE,NOM_FENETRE));
            return 0;
        }
#endif
        initcursor(hInstance);

        if (!hPrevInstance)
            if (!InitApplication(hInstance))
            {
                killcursor();
                return (FALSE);
            }


        if (!InitInstance(hInstance, nCmdShow))
        {
            killcursor();
            return (FALSE);
        }

        initwindow();

        strcpy(commandline,lpCmdLine);
        argc=get_args(argv,commandline);

        result=main(argc,argv);
	}

    /* meme en cas d'arret brutal on passera par ici */

#ifdef USER_FREE_FUNCTION
    free_all();
#endif
    /* pour liberer les zones allouees par VOTRE programme */

    return result;
}




























/***************************************************************/
/* gestion des 16 couleurs */

void setcolor(int c)
{
    if (nocolor!=c)
    {
        nocolor=c;  
        
        if (c==MASK)
            c=0;
        else
            if (c==FOND)
                c=255;
        
        cur_color=c;
    }
}



void getpalette(unsigned int *pal)
{
    int i;

    for(i=0;i<16;i++) pal[i]=oldpal[i];
}



void DDChangePalette(void)
{
    IDirectDrawPalette* ddpal;
    int                 i;
    PALETTEENTRY        ape[256];


#ifdef EXTENSION
    for (i=0; i<256; i++)
    {
        ape[i].peRed   = (BYTE)(oldpal256[i*3+0] >> 8);
        ape[i].peGreen = (BYTE)(oldpal256[i*3+1] >> 8);
        ape[i].peBlue  = (BYTE)(oldpal256[i*3+2] >> 8);
        ape[i].peFlags = (BYTE)0;
    }
#else
    for (i=0; i<16; i++)
    {
        ape[i].peRed   = (BYTE)(((oldpal[i] >> 8) & 0x0F) * 255 / 15);
        ape[i].peGreen = (BYTE)(((oldpal[i] >> 4) & 0x0F) * 255 / 15);
        ape[i].peBlue  = (BYTE)(((oldpal[i] >> 0) & 0x0F) * 255 / 15);
        ape[i].peFlags = (BYTE)0;
    }
#endif
    
	ddalert("SetEntries",lpPalette->lpVtbl->SetEntries(lpPalette,0,0,256,ape));

	return ddpal;
}


void setpalette(unsigned int *pal)
{
    unsigned int i,r,v,b;

    for(i=0;i<16;i++) oldpal[i]=pal[i];

#ifdef EXTENSION
    for(i=0;i<16;i++)
    {
        r=(pal[i]>>8)&15;
        v=(pal[i]>>4)&15;
        b=pal[i]&15;
        oldpal256[i*3+0]=r*4096;
        oldpal256[i*3+1]=v*4096;
        oldpal256[i*3+2]=b*4096;
    }
    setbigpalette(0,16,&oldpal256[0]);
#else
	DDChangePalette();
#endif
}







/***************************************************************/
/* gestion des 256 couleurs */

#ifdef EXTENSION        
void setbigcolor(int n,unsigned int r,unsigned int v,unsigned int b)
{
    unsigned int tmp[3];

    tmp[0]=r;
    tmp[1]=v;
    tmp[2]=b;

    setbigpalette(n,1,tmp);
}

void getbigcolor(int n,unsigned int *r,unsigned int *v,unsigned int *b)
{
    *r=oldpal256[n*3+0];
    *v=oldpal256[n*3+1];
    *b=oldpal256[n*3+2];
}

void setbigpalette(int no,int nb,unsigned int *ptr)
{
    int i;

    for(i=0;i<nb;i++)
    {
        oldpal256[(no+i)*3+0]=ptr[i*3+0];
        oldpal256[(no+i)*3+1]=ptr[i*3+1];
        oldpal256[(no+i)*3+2]=ptr[i*3+2];
    }

	DDChangePalette();
}

void getbigpalette(int no,int nb,unsigned int *ptr)
{
    int i;

    for(i=0;i<nb;i++)
    {
        ptr[i*3+0]=oldpal256[(no+i)*3+0];
        ptr[i*3+1]=oldpal256[(no+i)*3+1];
        ptr[i*3+2]=oldpal256[(no+i)*3+2];
    }
}
#endif











/************************************************************/
/* instructions de gestion des ecrans graphiques */





BOOL RestoreSurfaces( void )
{
    HRESULT     ddrval;
    
    ddrval = lpFrontBuffer->lpVtbl->Restore(lpFrontBuffer);
    if( ddrval != DD_OK )
	{
		//ddalert("Restore",ddrval);
        return FALSE;
	}

    return TRUE;
}



static void FlipScreen( void )
{
    HRESULT     ddrval;

    // Flip the surfaces
    while( 1 )
    {
        ddrval = lpFrontBuffer->lpVtbl->Flip( lpFrontBuffer, NULL, 0 );
		if( ddrval == DD_OK )
        {
            break;
        }
        if( ddrval == DDERR_SURFACELOST )
        {
            if( !RestoreSurfaces() )
            {
                break;
            }
        }
        if( ddrval != DDERR_WASSTILLDRAWING )
        {
            break;
        }
    }

}


void simplescreen(void)
{
    if (nbscreen==2)
    {
        nbscreen=1;
    }                                                  
}


void doublescreen(void)
{
    if (nbscreen==1)
    {
        nbscreen=2;
    }
}


void cls(void)
{
    register unsigned long *ptr;
    register long i;
    register long imax; 
	register long lcol;
    int t;

    xcur=ycur=0;
    t=nocolor;
    setcolor(0);
    imax=((long)BPL)*HEIGHT*RATIO/4;
    ptr=(unsigned long *)lockworkscreen();
	lcol=cur_color+(cur_color<<8)+(cur_color<<16)+(cur_color<<24);
    if (ptr)
	{
		if ((HEIGHT==200)||(HEIGHT==400))
		{
			ptr-=video_offset/4;
			imax=BPL*480/4;
		}
		for(i=0;i<imax;i++)
		    *(ptr++)=lcol;
	}
	releaseworkscreen();
    setcolor(t);
}


void copyscreen(void)
{
	void *ecran;

	initbloc(&ecran);
	swap();
	getbloc(&ecran,0,0,WIDTH,HEIGHT);
	swap();
	putbloc(&ecran,0,0);
	freebloc(&ecran);
}


void refresh(void)
{
    vblclock++;

	lastworkscreen=0;

	FlipScreen();
	copyscreen();
	RedrawWindow(window,NULL,NULL,RDW_INVALIDATE);
    buthit(); 

}

void swap(void)
{
    vblclock++;

	lastworkscreen=0;

	FlipScreen();
	RedrawWindow(window,NULL,NULL,RDW_INVALIDATE);
    buthit();                   
}









/******************************************************************/
/* instructions graphiques de base */
#ifdef BORLANDC
#undef sgn(x)
#undef abs(x)  
#endif
#define sgn(x) ((x==0)?(0):((x<0)?(-1):(1)))
#define abs(x) ((x<0)?(-(x)):(x))


void pellipse(int x,int y,int lx,int ly,int c)
{
    register int i,r,a,s;
    register int ry=(ly+1)>>1;
    register int rx=lx>>1;

    if (c>=0) setcolor(c);
    for(i=0,a=64;i<ry;i++)
    {
        s=(int)(((long)(ry-i)<<14L)/ry);
        while(sn[a]>s) a++;
        r=rx+(int)((cs[a]*(long)rx)>>14L);
        hline(x+r,y+i,x+lx-1-r,-1);
        hline(x+r,y+ly-i-1,x+lx-1-r,-1);
    }
}


void pbox(int x,int y,int lx,int ly,int c)
{
    register int i;

    if (c>=0) setcolor(c);
    for(i=y;i<y+ly;i++) hline(x,i,x+lx-1,-1);
}




void rline(int x1,int y1,int x2,int y2,int c)
{
    register long a,x,y;
    register int i;
    register int d;
    register int dx=x2-x1;
    register int dy=y2-y1;

    if (c>=0) setcolor(c);

    if (!dx) vline(x1,y1,y2,-1);
    else
        if (!dy) hline(x1,y1,x2,-1);
        else
        {
            if (abs(dx)>abs(dy))
            {
                d=sgn(dx);
                a=((long)dy<<16L)/abs(dx);
                for(i=x1,y=32768L+((long)y1<<16L);i!=x2;i+=d,y+=a)
                    plot(i,(int)(y>>16),-1);
            }
            else
            {
                d=sgn(dy);
                a=((long)dx<<16L)/abs(dy);
                for(i=y1,x=32768L+((long)x1<<16L);i!=y2;i+=d,x+=a) 
                    plot((int)(x>>16),i,-1);
            }

            plot(x2,y2,-1);
        }

}



void dline(int x1,int y1,int x2,int y2,int c)
{
    register long a,x,y;
    register int i;
    register int d;
    register int dx=x2-x1;
    register int dy=y2-y1;

    if (c>=0) setcolor(c);

    if (!dx) vline(x1,y1,y2,-1);
    else
        if (!dy) hline(x1,y1,x2,-1);
        else
        {
            if (abs(dx)>abs(dy))
            {
                d=sgn(dx);
                a=((long)dy<<16L)/abs(dx);
                for(i=x1,y=32768L+((long)y1<<16L);i!=x2;i+=d,y+=a)
                    plot(i,(int)(y>>16),-1);
            }
            else
            {
                d=sgn(dy);
                a=((long)dx<<16L)/abs(dy);
                for(i=y1,x=32768L+((long)x1<<16L);i!=y2;i+=d,x+=a) 
                    plot((int)(x>>16),i,-1);
            }

            plot(x2,y2,-1);
        }

}


void dbox(int x,int y,int lx,int ly,int c)
{
    if (c>=0) setcolor(c);
    if (!lx) lx++;
    if (!ly) ly++;
    hline(x,y,x+lx-1,-1);
    hline(x,y+ly-1,x+lx-1,-1);
    vline(x,y,y+ly-1,-1);
    vline(x+lx-1,y,y+ly-1,-1);
}





void plot(int x,int y,int c)
{
	register unsigned char *ptr;
	register unsigned char cur_c;

    if (c>=0) setcolor(c);

	cur_c=cur_color;

	ptr=lockworkscreen();
	if (ptr)
	{
		if (RATIO==1)
			*(ptr+((long)y)*BPL+x)=cur_c;
		else
		{
			ptr+=((long)y)*BPL*RATIO+x*RATIO;
			*ptr=cur_c; ptr++;
			*ptr=cur_c; ptr+=BPL;
			*ptr=cur_c; ptr--;
			*ptr=cur_c;
		}
		releaseworkscreen();
	}
}


void vline(int x,int y,int y1,int c)
{
    register unsigned char *ptr;
    register int dy;
    register int i;
    register unsigned char cur_c;

    if (c>=0) setcolor(c);

    ptr=lockworkscreen();
	if (!ptr) return;

	ptr+=((long)y)*BPL*RATIO+x*RATIO;

    dy=sgn(y1-y);

    cur_c=cur_color;
    if (!dy) dy++;

    for(i=y;i!=y1+dy;i+=dy,ptr+=dy*BPL*RATIO)
	{
		if (RATIO==1)
			*ptr=cur_c;
		else
		{
			*ptr=cur_c;
			*(ptr+1)=cur_c;
			*(ptr+BPL)=cur_c;
			*(ptr+1+BPL)=cur_c;
		}
	}

	releaseworkscreen();	
}


void hline(int x,int y,int x1,int c)
{
    register unsigned char *ptr;
    register int n=x1-x;
    register int i;
    register unsigned char cur_c;


    if (c>=0) setcolor(c);             

	ptr=lockworkscreen();
	if (!ptr) return;

	ptr+=((long)y)*BPL*RATIO+x*RATIO;

    cur_c=cur_color;
    if (n>=0)
	{
		if (RATIO==1)
		    for(i=0;i<=n;i++) *(ptr++)=cur_c;
		else
		    for(i=0;i<=n;i++)
			{
				*ptr=cur_c;
				*(ptr+1)=cur_c;
				*(ptr+BPL)=cur_c;
				*(ptr+1+BPL)=cur_c;
				ptr+=RATIO;
			}
	}
    else
    {
		if (RATIO==1)
		{	
			n= -n-1;
			*ptr=cur_c;
			for(i=0;i<=n;i++)
				*(--ptr)=cur_c;
		}
		else
		{
			n= -n-1;
			*ptr=cur_c;
			*(ptr+1)=cur_c;
			*(ptr+BPL)=cur_c;
			*(ptr+1+BPL)=cur_c;
			for(i=0;i<=n;i++)
			{
				ptr-=RATIO;
				*(ptr)=cur_c;
				*(ptr+1)=cur_c;
				*(ptr+BPL)=cur_c;
				*(ptr+1+BPL)=cur_c;								
			}
		}
    }       

	releaseworkscreen();
}



void polyline(int n,int *tp,int c)
{
    register int i;

    if (c>=0) setcolor(c);
    for(i=0;i<n-1;i++) rline(tp[i*2],tp[i*2+1],tp[i*2+2],tp[i*2+3],-1);
}


static void fplot(int x,int y)
{
    register int l=filltab[y];
    register int d;

    if (l>=0)
    {
        d=l-x;
        if ((!d)||(d==1)||(d== -1))
        {
            plot(x,y,-1);
            filltab[y]=x;
        }
        else
        {
            hline(x,y,l,-1);
            filltab[y]= -1;
        }
    }
    else
    {
        plot(x,y,-1);
        filltab[y]=x;
    }

}


static void fline(int x1,int y1,int x2,int y2)
{
    register long a,x,y;
    register int i;
    register int d;
    register int dx=x2-x1;
    register int dy=y2-y1;


    if (dx|dy)
        if (abs(dx)>abs(dy))
        {
            d=sgn(dx);
            a=((long)dy<<16L)/abs(dx);
            for(i=x1,y=32768L+((long)y1<<16L);i!=x2;i+=d,y+=a)
                fplot(i,(int)(y>>16));
        }
        else
        {
            d=sgn(dy);
            a=((long)dx<<16L)/abs(dy);
            for(i=y1,x=32768L+((long)x1<<16L);i!=y2;i+=d,x+=a) 
                fplot((int)(x>>16),i);
        }

    fplot(x2,y2);

}





void polyfill(int n,int *tp,int c)
{
    register int i;

    if (c>=0) setcolor(c);
    
    for(i=0;i<HEIGHT;i++) filltab[i]= -1;
    for(i=0;i<n-1;i++) fline(tp[i*2],tp[i*2+1],tp[i*2+2],tp[i*2+3]);
}












/*****************************************************************/
/* instructions d'allocation memoire                             */

long ramfree(void)
{
    return GetFreeSpace(NULL);
}
                      
void *memalloc(long size)
{     
#ifdef BORLANDC
    return((void *)farmalloc(size));              
#else
#ifdef WIN32                      
    return((void *)malloc(size));
#else
    return((void *)_fmalloc((size_t)size));
#endif
#endif
}

void memfree(void **ptr)
{               
#ifdef BORLANDC
    if (*ptr) farfree(*ptr);
#else
#ifdef WIN32
    if (*ptr) free(*ptr);
#else                          
    if (*ptr) _ffree(*ptr);
#endif
#endif
    *ptr=0L;
}









/**************************************************************/
/* transferts de blocs memoire */

void bmove(void *src,void *dst,long len)
{
    char huge *hs;
    char huge *hd;
    char *s;
    char *d;
    long hi;
    unsigned int i,l;

    if (len>65535L)
    {
        hs=(char huge *)src;
        hd=(char huge *)dst;
        if (hs>hd)
            for(hi=0L;hi<len;hi++) *(hd++)= *(hs++);
        else
            for(hd+=len,hs+=len,hi=0L;hi<len;hi++) *(--hd)= *(--hs);
    }
    else                                       
    {
        l=(unsigned int)len;
        s=(char *)src;
        d=(char *)dst;
        if (s>d)
            for(i=0;i<l;i++) *(d++)= *(s++);
        else
            for(d+=l,s+=l,i=0;i<l;i++) *(--d)= *(--s);
    }
}










/*****************************************************************/
/* instructions de gestion d'horloge                             */


unsigned long systime(void)
{
    return GetTickCount();
}


void waitdelay(unsigned int n)
{
    unsigned long start;
    start=systime();
    do
    {
        buthit();
    }
    while((unsigned long)(systime()-start)<((unsigned long)n));
}
















/******************************************************************/
/* fonctions aleatoires                                           */


int randint(void)
{
    return(rand()&0x7fff);
}

int randval(int n)
{
    return(randint()%n);
}















/**************************************************************/
/* gestion des fichiers */

#define MAXFILE 16
HMMIO ftab[MAXFILE];
char fused[MAXFILE]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

static int nextfile(void)
{
    int r=0;
    int i;

    for(i=0;i<MAXFILE;i++) if (!fused[i]) r=i+1;
    return(r);
}



static void verifnom(char *s,char *s2)
{
    while(*s)
    {
        *s2= *s;
        if (*s=='/') *s2='\\';
        s++;
        s2++;
    }
    *s2=0;
}


int bexist(char *nom)
{
    HMMIO fp;
    char nom2[200];
    verifnom(nom,nom2);

    fp = mmioOpen(nom2,NULL,MMIO_READ);
    if (fp)
    {
        mmioClose(fp,NULL);
        return(1);
    }
    else
        return(0);
}


long bsize(char *nom)
{
    HMMIO fp;
    char nom2[200];
    long size=0L;

    verifnom(nom,nom2);

    fp = mmioOpen(nom2,NULL,MMIO_READ);
    if (fp)
    {
        size=mmioSeek(fp, 0L, SEEK_END);
        mmioClose(fp,NULL);
    }

    return(size);
}



int bsave(char *nom,void *adr,long offset,long len)
{
    void *buffer;
    long filesize;
    HMMIO fp;
    int r= -1;

    char nom2[200];
    verifnom(nom,nom2);
    fp = mmioOpen(nom2,NULL,MMIO_READ);
    if (fp)
    {
        r=0;
        filesize = mmioSeek(fp, 0L, SEEK_END);
        mmioClose(fp,NULL);

        if (!r)
        {
            r= -1;
            buffer=memalloc(filesize);
            if (buffer)
            {
                if (!bload(nom,buffer,0L,filesize))
                {
                    bmove(adr,((char *)buffer)+offset,len);
                    r=bmake(nom,buffer,filesize);
                }
                memfree((void **)&buffer);
            }
        }
    }

    return(r);
}





int bload(char *nom,void *adr,long offset,long len)
{
    HMMIO fp;
    int r= -1;
    char nom2[200];

    verifnom(nom,nom2);

    fp = mmioOpen(nom2,NULL,MMIO_READ);
    if (fp)
    {
        mmioSeek(fp, offset, SEEK_SET);
        if (len==mmioRead(fp,adr,len))
            r=0;
        else
            r= -1;
        mmioClose(fp,NULL);
    }

    return(r);
}




int bmake(char *nom,void *adr,long len)
{
    HMMIO fp;
    int r= -1;
    char nom2[200];

    verifnom(nom,nom2);

    fp = mmioOpen(nom2,NULL,MMIO_WRITE|MMIO_CREATE);
    if (fp)
    {
        if (len==mmioWrite(fp,adr,len))
            r=0;
        else
            r= -1;
        mmioClose(fp,NULL);
    }
    return(r);
}



int bcreate(char *nom)
{
    HMMIO fp;
    int r;
    char nom2[200];

    r=nextfile();
    if (r)
    {
        verifnom(nom,nom2);
        fp = mmioOpen(nom2,NULL,MMIO_WRITE|MMIO_CREATE);
        if (fp) { ftab[r-1]=fp; fused[r-1]=1; }
        else r=0;
    }
    return(r);
}



int bopen(char *nom)
{
    HMMIO fp;
    int r;
    char nom2[200];

    r=nextfile();
    if (r)
    {
        verifnom(nom,nom2);
        fp = mmioOpen(nom2,NULL,MMIO_READ);
        if (fp) { ftab[r-1]=fp; fused[r-1]=1; }
        else r=0;
    }
    return(r);
}



int bclose(int n)
{
    HMMIO fp;
    int r=0;

    if (n)
    {
        fp=ftab[n-1];
        mmioClose(fp,NULL);
        fused[n-1]=0;
    }
    return(r);
}




int bwrite(int n,void *adr,long len)
{
    HMMIO fp;
    int r= -1;

    if (n)
    {
        fp=ftab[n-1];
        if (fp)
        {
            if (len==mmioWrite(fp,adr,len))
                r=0;
            else
                r= -1;
        }
    }
    return(r);
}




int bread(int n,void *adr,long len)
{
    HMMIO fp;
    int r= -1;

    if (n)
    {
        fp=ftab[n-1];
        if (fp)
        {
            if (len==mmioRead(fp,adr,len))
                r=0;
            else
                r= -1;
        }
    }
    return(r);
}


















/**************************************************************/
/* gestion de la souris et du clavier */

void hide(void)
{
    realmsemode--;
    if (realmsemode==0)
        SetCursor(invisible_cursor);
}

void show(void)
{
    realmsemode++;
    if (realmsemode==1)
        SetCursor(visible_cursor);
}




static void addchar(int c)
{
    if (keybufnbr<256)
    {
        keybuffer[keybufptr++]=c;
        keybufnbr++;
        keybufptr&=255;
    }
}



static void addbut(int k,int x,int y,int c)
{
    int ptr;

    if (nbrbuffer<32)
    {
        ptr=(msebufptr>>1);
        msebuffer[ptr]=k;
        msebuffer[ptr+1]=x;
        msebuffer[ptr+2]=y;
        msebuffer[ptr+3]=c;
        msebufptr+=8;
        msebufptr&=255;
        nbrbuffer++;
    }
}

static void msemoved(int x,int y,unsigned int k)
{
    mousex=x;
    mousey=y;
    if (mousex<0) mousex=0;
    if (mousey<0) mousey=0;
    if (mousex>(WIDTH-1)) mousex=(WIDTH-1);
    if (mousey>(HEIGHT-1)) mousey=(HEIGHT-1);
    mousek=0;
    if (k&MK_LBUTTON) mousek|=1;
    if (k&MK_RBUTTON) mousek|=2;
}


static unsigned int scancode(unsigned int k)
{

    switch(k)
    {
        case 0xDB: k=')'; break;
        case 0xBB: k='='; break;
        case 0xDD: k='^'; break;
        case 0xBA: k='$'; break;
        case 0xC0: k='%'; break;
        case 0xDC: k='*'; break;
        case 0xBC: k=','; break;
        case 0xBE: k=';'; break;
        case 0xBF: k=':'; break;
        case 0xDF: k='!'; break;
        case 111: k='/'; break;
        case 106: k='*'; break;
        case 109: k='-'; break;
        case 107: k='+'; break;
        case 110: k='.'; break;
        case 96: k='0'; break;
        case 97: k='1'; break;
        case 98: k='2'; break;
        case 99: k='3'; break;
        case 100: k='4'; break;
        case 101: k='5'; break;
        case 102: k='6'; break;
        case 103: k='7'; break;
        case 104: k='8'; break;
        case 105: k='9'; break;
        case 0x26: k='^'; break;
        case 0x28: k='v'; break;
        case 0x25: k='<'; break;
        case 0x27: k='>'; break;
        case 226: k='<'; break;
    }
                         
    
    if ((k>='A')&&(k<='Z')) k+=32;

    if (k>127) k=0;
    if ((k<32)&&(k!=27)&&(k!=13)&&(k!=8)) k=0;

    if (k=='^') k=0;
    
    return k;
}


static int lookevent(MSG *msg)
{
    int x,y,r;
    unsigned int k;
    long c;

    k=((unsigned int)msg->wParam)&255;
    x=((int)LOWORD(msg->lParam))/RATIO;
    y=(((int)HIWORD(msg->lParam))-y_offset)/RATIO;
    r=1;
    if ((x<0)||(x>10000)) x=0;
    if ((y<0)||(y>10000)) y=0;
    if (x>=WIDTH) x=WIDTH-1;
    if (y>=HEIGHT) y=HEIGHT-1;

    switch (msg->message)
    {
        case WM_MOUSEMOVE:
            realmsex=x;
            realmsey=y;
            msemoved(x,y,k);
            break;

        case WM_LBUTTONDOWN:
            msemoved(x,y,k);

            if ((x<WIDTH)&&(y<HEIGHT)&&(x>=0)&&(y>=0))
                addbut(1,x,y,(int)(msg->time&65535L));

            break;

        case WM_LBUTTONUP:
            msemoved(x,y,k);

            if ((x<WIDTH)&&(y<HEIGHT)&&(x>=0)&&(y>=0))
                addbut(5,x,y,(int)(msg->time&65535L));

            break;

        case WM_RBUTTONDOWN:
            msemoved(x,y,k);

            if ((x<WIDTH)&&(y<HEIGHT)&&(x>=0)&&(y>=0))
                addbut(2,x,y,(int)(msg->time&65535L));

            break;

        case WM_RBUTTONUP:
            msemoved(x,y,k);

            if ((x<WIDTH)&&(y<HEIGHT)&&(x>=0)&&(y>=0))
                addbut(6,x,y,(int)(msg->time&65535L));

            break;


        case WM_KEYUP:
            k&=255; 
            keystate[k]=0;
                k=scancode(k);
            if (k) keymap[k]=0;
            break;

        case WM_KEYDOWN:
            k&=255;

            keystate[k]=-128; 
            k=scancode(k);
            if (ToAscii((UINT)msg->wParam,(UINT)((msg->lParam>>16)&255),(BYTE FAR*)keystate,(DWORD FAR*)&c,0))
            {
                c&=255;
                if (c!='^') addchar((int)c);
            }
            else
            {
                if (k==0x14)
                    keystate[0x10]^=128;
                if ((k)&&(!(msg->lParam>>30)))
                                addchar(k);
            }


            if ((k)&&(!(msg->lParam>>30)))
            {
                keymap[k]=1;
            }
            break;

                                

        default:
            r=0;
                break;
    }

    return r;
}

                                                



int getmouse(void)
{
    int ptr;

    buthit();

    if (nbrbuffer)
    {
        ptr=(lstbufptr>>1);
        msek=msebuffer[ptr];
        msex=msebuffer[ptr+1];
        msey=msebuffer[ptr+2];
        mclk=msebuffer[ptr+3];
        lstbufptr+=8;
        lstbufptr&=255;
        nbrbuffer-=1;

        /* msek: bit 0:bouton gauche  bit 1:bouton droit  bit 2:0:presse 1:relache */

        return msek;
    }

    return 0;
}







int buthit(void)
{
    MSG msg;

    while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
        if (!lookevent(&msg))
        {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }

    return(nbrbuffer);
}





void empty(void)
{
    kbhit();
    buthit();

    keybufnbr=keybufptr=keybufend=0;
    nbrbuffer=lstbufptr=msebufptr=0;
}

                                      


char getch(void)
{
    char c=0;
    MSG msg;

    while(!kbhit())
    {
        if (GetMessage(&msg, NULL, 0, 0))
            if (!lookevent(&msg))
            {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
            }
    }

    if (keybufnbr)
    {
        c=(char)keybuffer[keybufend++];
        keybufend&=255;
        keybufnbr--;
    }
    
    if (c=='') c='e';
    if (c=='') c='e';
    if (c=='') c='e';
    if (c=='') c='e';
    if (c=='') c='a';
    if (c=='') c='a';
    if (c=='') c='a';
    if (c=='') c='i';
    if (c=='') c='i';
    if (c=='') c='u';   
    if (c=='') c='u';
    if (c=='') c='u';
    if (c=='') c='o';
    if (c=='') c='o';
    
    return(c);
}



int kbhit(void)
{
    MSG msg;

    while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
        if (!lookevent(&msg))
        {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }

    return(keybufnbr);
}


int keyhit(void)
{
    int r;

    r=buthit();
    if (!r) r=kbhit();
    return(r);
}



void confirm(void)
{
    MSG msg;

    empty();           
    while(getmouse()!=1)
    {
        if (GetMessage(&msg, NULL, 0, 0))
            if (!lookevent(&msg))
            {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
            }       
    }
}












/***************************************************************/
/* gestion de blocs graphiques */



typedef struct
{
    int lx;
    int ly;
    int id;
    unsigned char *bloc;
} 
pixbloc;


void initbloc(void **blocptr)
{
    pixbloc *bloc;
    *blocptr=memalloc((long)sizeof(pixbloc));
    bloc=(pixbloc *)*blocptr;
    if (*blocptr) bloc->id=0;
}


void freebloc(void **blocptr)
{
    pixbloc *bloc;
    if (*blocptr)
    {
        bloc=(pixbloc *)*blocptr;
        if (bloc->id) memfree((void **)&bloc->bloc);
        bloc->id=0;
        memfree(blocptr);
    }
}



void getbloc(void **blocptr,int x,int y,int lx,int ly)
{
    register int i,j,imax,jmax;
    register unsigned char *matrice,*matrice2;
    register long recal;
	unsigned char *ptr;

    pixbloc *bloc;

	ptr=lockworkscreen();
	if (!ptr) return;

    if (*blocptr)
    {
        bloc=(pixbloc *)*blocptr;
        if ((!bloc->id)||(bloc->lx!=lx)||(bloc->ly!=ly))
        {
            if (bloc->id) memfree((void **)&bloc->bloc);
            bloc->id=0;
            bloc->bloc=(unsigned char *)memalloc(((long)ly)*lx);
            if (bloc->bloc)
            {
                bloc->id=1;
                bloc->lx=lx;
                bloc->ly=ly;
            }
        }


        if (bloc->id)
        {
			if (RATIO==1)
			{
				recal=(long)(BPL-lx);
				matrice=ptr+x+((long)y)*BPL;
				matrice2=bloc->bloc;
				imax=ly;
				jmax=lx;

				for(i=0;i<imax;i++)
				{
					for(j=0;j<jmax;j++)
					{
						*(matrice2++)=*(matrice++);
					}
					matrice+=recal;
				}
			}
			else
			{
				recal=(long)(BPL*RATIO-lx*RATIO);
				matrice=ptr+x*RATIO+((long)y)*BPL*RATIO;
				matrice2=bloc->bloc;
				imax=ly;
				jmax=lx;    

				for(i=0;i<imax;i++)
				{
					for(j=0;j<jmax;j++)
					{
						*(matrice2++)=*matrice;
						matrice+=RATIO;
					}
					matrice+=recal;
				}
			}
        }
    }

	releaseworkscreen();
}



void copybloc(void **blocptrs,int xs,int ys,int lxs,int lys,
                    void **blocptrd,int xd,int yd)
{                                     
    register int i,j,imax,jmax;
    register unsigned char *matrice,*matrice2;
    register long recal,recal2;
    pixbloc *blocs;
    pixbloc *blocd;
    if ((*blocptrs)&&(*blocptrd))
    {
        blocs=(pixbloc *)*blocptrs;
        blocd=(pixbloc *)*blocptrd;
        if ((blocs->id)&&(blocd->id))
        if ((xd>=0)&&(yd>=0)&&(xs>=0)&&(ys>=0))
        if ((xs+lxs<=blocs->lx)&&(ys+lys<=blocs->ly))
        if ((xd+lxs<=blocd->lx)&&(yd+lys<=blocd->ly))
        {
            matrice=blocs->bloc+xs+ys*blocs->lx;
            matrice2=blocd->bloc+xd+yd*blocd->lx;
            recal=(long)(blocs->lx-lxs);
            recal2=(long)(blocd->lx-lxs);
            imax=lys;
            jmax=lxs;

            for(i=0;i<imax;i++)
            {
                for(j=0;j<jmax;j++)
                {
                    *(matrice2++)=*(matrice++);
                }
                matrice+=recal;
                matrice2+=recal2;
            }
        }
    }
}




void putpbloc(void **blocptr,int x,int y,int xs,int ys,
                    int lxs,int lys)
{                           
    register i,j,imax,jmax;
    register unsigned char c;
    register unsigned char *matrice,*matrice2;
    register long recal,recal2;
	unsigned char *ptr;
    pixbloc *bloc;

	ptr=lockworkscreen();
	if (!ptr) return;

    if (*blocptr)
    {
        bloc=(pixbloc *)*blocptr;

        if ((y>=0)||(y+lys<=HEIGHT))
        {
            if (y<0)
            {
                ys+=(-y);
                lys-=(-y);
                y=0;
            }
            else
            if (y+lys>HEIGHT)
            {
                lys=HEIGHT-y;
            }
        }
        if (lys>0)
        if ((x>=0)&&(y>=0)&&(xs>=0)&&(ys>=0))
        if ((x<WIDTH)&&(y<HEIGHT))
        if ((x+lxs<=WIDTH)&&(y+lys<=HEIGHT))
        if (bloc->id)
        {
			if (RATIO==1)
			{
            
				recal2=(long)(BPL-lxs);

				matrice=bloc->bloc+xs+ys*bloc->lx;
				matrice2=ptr+x+((long)y)*BPL;
				recal=(long)(bloc->lx-lxs);
				imax=lys;
				jmax=lxs;

				if (bloc->id==1)
					for(i=0;i<imax;i++)
					{
						for(j=0;j<jmax;j++)
						{
							*(matrice2++)=*(matrice++);
						}
						matrice+=recal;
						matrice2+=recal2;
					}
				else   
					for(i=0;i<imax;i++)
					{
						for(j=0;j<jmax;j++)
						{
							c=*(matrice++);
							if (c!=255)                                         
								*matrice2=c;
							matrice2++;
						}
						matrice+=recal;
						matrice2+=recal2;
					}
			}
			else
			{
				            
				recal2=(long)(BPL*RATIO-lxs*RATIO);

				matrice=bloc->bloc+xs+ys*bloc->lx;
				matrice2=ptr+x*RATIO+((long)y)*BPL*RATIO;
				recal=(long)(bloc->lx-lxs);
				imax=lys;
				jmax=lxs;

				if (bloc->id==1)
					for(i=0;i<imax;i++)
					{
						for(j=0;j<jmax;j++)
						{
							c=*(matrice++);
							*matrice2=c;
							*(matrice2+1)=c;
							*(matrice2+BPL)=c;
							*(matrice2+1+BPL)=c;
							matrice2+=RATIO;
						}
						matrice+=recal;
						matrice2+=recal2;
					}
				else   
					for(i=0;i<imax;i++)
					{
						for(j=0;j<jmax;j++)
						{
							c=*(matrice++);
							if (c!=255)                                         
							{
								*matrice2=c;
								*(matrice2+1)=c;
								*(matrice2+BPL)=c;
								*(matrice2+1+BPL)=c;
							}
							matrice2+=RATIO;
						}
						matrice+=recal;
						matrice2+=recal2;
					}

			}
                            
        }
    }

	releaseworkscreen();
}



void putbloc(void **blocptr,int x,int y)
{
    pixbloc *bloc;
    if (*blocptr)
    {
        bloc=(pixbloc *)*blocptr;
        putpbloc(blocptr,x,y,0,0,bloc->lx,bloc->ly);
    }
}


void getmask(void **blocptr,int x,int y)
{
    pixbloc *bloc;
    if (*blocptr)
    {
        bloc=(pixbloc *)*blocptr;
        if (bloc->id)
            bloc->id=2;
    }
    
    x+=0;
    y+=0;
}

















/*****************************************************************/
/* instructions de gestion de fonte */

void affchar(int x,int y,char c)
{
	register unsigned char *ptr;
    register unsigned int n,i,k,m,f,t;

	ptr=lockworkscreen();
	if (!ptr) return;

    t=nocolor;
    if ((c>31)&&((unsigned char)c<=127))
    {
        if (c>95) c-=(char)32;
    
        n=c-32;
    
		ptr+=y*BPL*RATIO+x*RATIO;

		if (RATIO==1)
			for(i=0;i<8;i++)
			{
				for(m=sysfonte[n*16+i],f=sysfonte[n*16+i+8],k=0;k<8;k++,f<<=1,m<<=1)
				{
					if (!(m&128))
						if (f&128)
							*ptr=t;
						else
							*ptr=0;
					ptr++;
				}
				ptr+=BPL-8;
			}
		else
			for(i=0;i<8;i++)
			{
				for(m=sysfonte[n*16+i],f=sysfonte[n*16+i+8],k=0;k<8;k++,f<<=1,m<<=1)
				{
					if (!(m&128))
						if (f&128)
						{
							*ptr=t;
							*(ptr+1)=t;
							*(ptr+BPL)=t;
							*(ptr+1+BPL)=t;
						}
						else
						{
							*ptr=0;
							*(ptr+1)=0;
							*(ptr+BPL)=0;
							*(ptr+1+BPL)=0;
						}
					ptr+=RATIO;
				}
				ptr+=BPL*RATIO-8*RATIO;
			}
    
        setcolor(t);
    }

	releaseworkscreen();
}



void afftext(int x,int y,char *s)
{
    while(*s)
    {
        affchar(x,y,*s++);
        x+=8;
        if (x>WIDTH-8) { x=0; y+=8; if (y>(HEIGHT-8)) y=0; }
    }
}


void printchar(char c)
{
    int t;
	unsigned char *ptr;

    while(ycur>(HEIGHT-8))
    {
        ycur-=8;       
		ptr=lockworkscreen();
        if (ptr)
		{
			bmove((void *)(ptr+8L*BPL*RATIO),(void *)ptr,(HEIGHT-8)*((long)BPL*RATIO));
			releaseworkscreen();
		}

        t=nocolor;
        pbox(0,(HEIGHT-8),WIDTH,8,0);
        setcolor(t);
    }


    if (c==10) xcur= -8;
    else
    if (c==13) { ycur+=8; xcur= -8; }
    else
    affchar(xcur,ycur,c);

    xcur+=8;
    if (xcur>(WIDTH-8)) { xcur=0; ycur+=8; }


}


void print(char *s)
{
    while(*s) printchar(*s++);

    xcur=0;
    ycur+=8;
}















/*******************************************************************/
/* gestion des samples */

                                            

void playsample(void *adr,long len,long freq)
{                      
	int n;
	int i;
	PWAVEFORMATEX p;

	if (!bSoundEnabled) return;

	
	n=(int)*((char *)adr);
	if ((n>=0)&&(n<MAXSAMPLE))
	{
		if ((freq!=splfrq[n])||(len!=spllen[n]))
		{        

			p=(PWAVEFORMATEX)(((char*)adr)+2);

			p->wFormatTag=1;
			p->nChannels=1;
			p->nSamplesPerSec=freq;
			p->nAvgBytesPerSec=freq;
			p->nBlockAlign=1;
			p->wBitsPerSample=8;
			p->cbSize=len-(2+sizeof(WAVEFORMATEX));
                            
			                
			splfrq[n]=freq; 
			spllen[n]=len;               

			for(i=0;i<hsoSample[n]->iAlloc;i++)
			{
				IDirectSoundBuffer_SetFrequency(hsoSample[n]->Buffers[i],freq);
			}


		}                      
	    
		SndObjPlay(hsoSample[n], 0);
	}                    
                                                                   
}



void stopsample()
{
	char command[80];
	int i;

	if (!bSoundEnabled) return;

	for(i=0;i<MAXSAMPLE;i++)
	if (splfrq[i])
	{	
		SndObjStop(hsoSample[i]);
	}
}
                           
                           
                           
void fixsample(void *adr,long len)
{                                          
	int i;  
	PWAVEFORMATEX p;

	if (!bSoundEnabled) return;


	if (len<=sizeof(WAVEFORMATEX)+2) 
	{
		*((char *)adr)=-1;
		return;
	}

	i=0;
	while ((splfrq[i]) && (i<MAXSAMPLE)) i++;
	if (i<MAXSAMPLE)
	{
		p=(PWAVEFORMATEX)(((char*)adr)+2);

		p->wFormatTag=1;
		p->nChannels=1;
		p->nSamplesPerSec=11025;
		p->nAvgBytesPerSec=11025;
		p->nBlockAlign=1;
		p->wBitsPerSample=8;
		p->cbSize=len-(2+sizeof(WAVEFORMATEX));
                            
		splfrq[i]=11025L;
		spllen[i]=len;
		
		*((char *)adr)=i;

        hsoSample[i] = SndObjCreateFromMemory(lpDS, 1, p, ((char *)adr)+2+sizeof(WAVEFORMATEX), len-(2+sizeof(WAVEFORMATEX)));

	}
	            
}



                              